import { Query, SelectableFromShape, CreateMethodsNames, AddQueryDefaults, SetQueryReturnsOne, SetQueryReturnsOneOptional, RelationJoinQuery, UpdateData, WhereArg, CreateData, JoinQueryMethod, DeleteMethodsNames, Db, IsolationLevel, TransactionOptions, Adapter, FromArg, FromResult, AdapterOptions, DbSharedOptions, RelationQuery, ShapeColumnPrimaryKeys, ShapeUniqueColumns, TableDataItemsUniqueColumns, TableDataItemsUniqueColumnTuples, UniqueConstraints, TableDataItemsUniqueConstraints, MapTableScopesOption, TableDataItem, ComputedColumnsBase, QueryData, QueryBase, TableDataFn, DbTableOptionScopes, SqlMethod, DefaultSchemaConfig, DefaultColumnTypes, RelationQueryBase, QueryBeforeHook, QueryAfterHook, AfterHook, WhereResult, MergeQuery } from 'pqb';
export * from 'pqb';
import { ColumnsShapeBase, EmptyObject, MaybeArray, RecordUnknown, CoreQueryScopes, ColumnShapeOutput, ColumnShapeInput, ColumnShapeInputPartial, ColumnSchemaConfig, QueryReturnType } from 'orchid-core';
export * from 'orchid-core';

interface RelationCommonOptions<Related extends TableClass = TableClass, Scope extends Query = Query> {
    scope?: ScopeFn<Related, Scope>;
    required?: boolean;
}
interface RelationRefsOptions<Column extends PropertyKey = string, Ref extends PropertyKey = string> {
    columns: Column[];
    references: Ref[];
}
interface RelationKeysOptions<PK extends PropertyKey = string, FK extends PropertyKey = string> {
    primaryKey: PK;
    foreignKey: FK;
}
type RelationRefsOrKeysOptions<Column extends PropertyKey = string, Ref extends PropertyKey = string, PK extends PropertyKey = string, FK extends PropertyKey = string> = RelationRefsOptions<Column, Ref> | RelationKeysOptions<PK, FK>;
interface RelationThroughOptions<Through extends PropertyKey = string, Source extends PropertyKey = string> {
    through: Through;
    source: Source;
}
type RelationHasOptions<Column extends PropertyKey = string, Ref extends PropertyKey = string> = RelationRefsOrKeysOptions<Column, Ref, Column, Ref>;

interface HasOne extends RelationThunkBase {
    type: 'hasOne';
    options: HasOneOptions;
}
type HasOneOptions<Columns extends ColumnsShapeBase = ColumnsShapeBase, Related extends TableClass = TableClass, Scope extends Query = Query, Through extends string = string, Source extends string = string> = RelationCommonOptions<Related, Scope> & (RelationHasOptions<keyof Columns, keyof InstanceType<Related>['columns']['shape']> | RelationThroughOptions<Through, Source>);
type HasOneParams<T extends RelationConfigSelf, Relation extends RelationThunkBase> = Relation['options'] extends RelationRefsOptions ? {
    [Name in Relation['options']['columns'][number]]: T['columns']['shape'][Name]['type'];
} : Relation['options'] extends RelationKeysOptions ? Record<Relation['options']['primaryKey'], T['columns']['shape'][Relation['options']['primaryKey']]['type']> : Relation['options'] extends RelationThroughOptions ? RelationConfigParams<T, T['relations'][Relation['options']['through']]> : never;
interface HasOneInfo<T extends RelationConfigSelf, Name extends keyof T['relations'] & string, TableQuery extends Query, Populate = T['relations'][Name]['options'] extends RelationRefsOptions ? Record<T['relations'][Name]['options']['references'][number], true> : T['relations'][Name]['options'] extends RelationKeysOptions ? Record<T['relations'][Name]['options']['foreignKey'], true> : never, Q extends Query = T['relations'][Name]['options'] extends RelationThroughOptions ? {
    [K in keyof TableQuery]: K extends 'meta' ? Omit<TableQuery['meta'], 'selectable'> & {
        as: Name;
        defaults: Populate;
        hasWhere: true;
        selectable: SelectableFromShape<TableQuery['shape'], Name>;
    } : K extends 'join' ? RelJoin : K extends CreateMethodsNames ? never : TableQuery[K];
} : {
    [K in keyof TableQuery]: K extends 'meta' ? Omit<TableQuery['meta'], 'selectable'> & {
        as: Name;
        defaults: Populate;
        hasWhere: true;
        selectable: SelectableFromShape<TableQuery['shape'], Name>;
    } : K extends 'join' ? RelJoin : TableQuery[K];
}, NestedCreateQuery extends Query = T['relations'][Name]['options'] extends RelationThroughOptions ? Q : AddQueryDefaults<Q, Populate>> {
    query: Q;
    methodQuery: T['relations'][Name]['options']['required'] extends true ? SetQueryReturnsOne<Q> : SetQueryReturnsOneOptional<Q>;
    joinQuery: RelationJoinQuery;
    one: true;
    omitForeignKeyInCreate: never;
    optionalDataForCreate: T['relations'][Name]['options'] extends RelationThroughOptions ? EmptyObject : {
        [P in Name]?: RelationToOneDataForCreate<{
            nestedCreateQuery: NestedCreateQuery;
            table: Q;
        }>;
    };
    dataForUpdate: {
        disconnect: boolean;
    } | {
        delete: boolean;
    } | {
        update: UpdateData<Q>;
    };
    dataForUpdateOne: {
        set: WhereArg<Q>;
    } | {
        upsert: {
            update: UpdateData<Q>;
            create: CreateData<NestedCreateQuery> | (() => CreateData<NestedCreateQuery>);
        };
    } | {
        create: CreateData<NestedCreateQuery>;
    };
    params: HasOneParams<T, T['relations'][Name]>;
}

interface HasMany extends RelationThunkBase {
    type: 'hasMany';
    options: HasManyOptions;
}
type HasManyOptions<Columns extends ColumnsShapeBase = ColumnsShapeBase, Related extends TableClass = TableClass, Scope extends Query = Query, Through extends string = string, Source extends string = string> = HasOneOptions<Columns, Related, Scope, Through, Source>;
type HasManyParams<T extends RelationConfigSelf, Relation extends RelationThunkBase> = Relation['options'] extends RelationRefsOptions ? {
    [Name in Relation['options']['columns'][number]]: T['columns']['shape'][Name]['type'];
} : Relation['options'] extends RelationKeysOptions ? Record<Relation['options']['primaryKey'], T['columns']['shape'][Relation['options']['primaryKey']]['type']> : Relation['options'] extends RelationThroughOptions ? RelationConfigParams<T, T['relations'][Relation['options']['through']]> : never;
interface HasManyInfo<T extends RelationConfigSelf, Name extends keyof T['relations'] & string, TableQuery extends Query, Populate extends PropertyKey = T['relations'][Name]['options'] extends RelationRefsOptions ? T['relations'][Name]['options']['references'][number] : T['relations'][Name]['options'] extends RelationKeysOptions ? T['relations'][Name]['options']['foreignKey'] : never, Q extends Query = T['relations'][Name]['options'] extends RelationThroughOptions ? {
    [K in keyof TableQuery]: K extends 'meta' ? Omit<TableQuery['meta'], 'selectable'> & {
        as: Name;
        defaults: Record<Populate, true>;
        hasWhere: true;
        selectable: SelectableFromShape<TableQuery['shape'], Name>;
    } : K extends 'join' ? RelJoin : K extends CreateMethodsNames ? never : TableQuery[K];
} : {
    [K in keyof TableQuery]: K extends 'meta' ? Omit<TableQuery['meta'], 'selectable'> & {
        as: Name;
        defaults: Record<Populate, true>;
        hasWhere: true;
        selectable: SelectableFromShape<TableQuery['shape'], Name>;
    } : K extends 'join' ? RelJoin : TableQuery[K];
}> {
    query: Q;
    methodQuery: Q;
    joinQuery: RelationJoinQuery;
    one: false;
    omitForeignKeyInCreate: never;
    optionalDataForCreate: {
        [P in Name]?: T['relations'][Name]['options'] extends RelationThroughOptions ? EmptyObject : RelationToManyDataForCreate<{
            nestedCreateQuery: T['relations'][Name]['options'] extends RelationThroughOptions ? Q : AddQueryDefaults<Q, Record<Populate, true>>;
            table: Q;
        }>;
    };
    dataForUpdate: {
        disconnect?: MaybeArray<WhereArg<Q>>;
        delete?: MaybeArray<WhereArg<Q>>;
        update?: {
            where: MaybeArray<WhereArg<Q>>;
            data: UpdateData<Q>;
        };
    };
    dataForUpdateOne: {
        set?: MaybeArray<WhereArg<Q>>;
        create?: CreateData<T['relations'][Name]['options'] extends RelationThroughOptions ? Q : AddQueryDefaults<Q, Record<Populate, true>>>[];
    };
    params: HasManyParams<T, T['relations'][Name]>;
}

type RelJoin = JoinQueryMethod & (<T extends Query>(this: T) => T);

interface BelongsTo extends RelationThunkBase {
    type: 'belongsTo';
    options: BelongsToOptions;
}
type BelongsToOptions<Columns extends ColumnsShapeBase = ColumnsShapeBase, Related extends TableClass = TableClass, Scope extends Query = Query> = RelationCommonOptions<Related, Scope> & RelationRefsOrKeysOptions<keyof Columns, keyof InstanceType<Related>['columns']['shape'], keyof InstanceType<Related>['columns']['shape'], keyof Columns>;
type BelongsToFKey<Relation extends RelationThunkBase> = Relation['options'] extends RelationRefsOptions ? Relation['options']['columns'][number] : Relation['options'] extends RelationKeysOptions ? Relation['options']['foreignKey'] : never;
type BelongsToParams<T extends RelationConfigSelf, Relation extends RelationThunkBase> = {
    [Name in BelongsToFKey<Relation>]: T['columns']['shape'][Name]['type'];
};
interface BelongsToInfo<T extends RelationConfigSelf, Name extends keyof T['relations'] & string, TableQuery extends Query, FK extends string, Required, Q extends Query = {
    [P in keyof TableQuery]: P extends 'meta' ? // Omit is optimal
    Omit<TableQuery['meta'], 'selectable'> & {
        as: Name;
        hasWhere: true;
        selectable: SelectableFromShape<TableQuery['shape'], Name>;
    } : P extends 'join' ? RelJoin : P extends CreateMethodsNames | DeleteMethodsNames ? never : TableQuery[P];
}> {
    query: Q;
    methodQuery: Required extends true ? SetQueryReturnsOne<Q> : SetQueryReturnsOneOptional<Q>;
    joinQuery: RelationJoinQuery;
    one: true;
    omitForeignKeyInCreate: FK;
    dataForCreate: {
        columns: {
            [L in FK]: T['columns']['shape'][L]['inputType'];
        };
        nested: Required extends true ? {
            [Key in Name]: RelationToOneDataForCreateSameQuery<Q>;
        } : {
            [Key in Name]?: RelationToOneDataForCreateSameQuery<Q>;
        };
    };
    optionalDataForCreate: EmptyObject;
    dataForUpdate: {
        disconnect: boolean;
    } | {
        set: WhereArg<Q>;
    } | {
        delete: boolean;
    } | {
        update: UpdateData<Q>;
    } | {
        create: CreateData<Q>;
    };
    dataForUpdateOne: {
        upsert: {
            update: UpdateData<Q>;
            create: CreateData<Q> | (() => CreateData<Q>);
        };
    };
    params: BelongsToParams<T, T['relations'][Name]>;
}

declare function transaction<T extends {
    $queryBuilder: Db;
}, Result>(this: T, fn: () => Promise<Result>): Promise<Result>;
declare function transaction<T extends {
    $queryBuilder: Db;
}, Result>(this: T, options: IsolationLevel | TransactionOptions, fn: () => Promise<Result>): Promise<Result>;

type OrchidORM<T extends TableClasses = TableClasses> = {
    [K in keyof T]: DbTable<InstanceType<T[K]>>;
} & {
    $transaction: typeof transaction;
    $adapter: Adapter;
    $queryBuilder: Db;
    /**
     * Use `$query` to perform raw SQL queries.
     *
     * ```ts
     * const value = 1;
     *
     * // it is safe to interpolate inside the backticks (``):
     * const result = await db.$query<{ one: number }>`SELECT ${value} AS one`;
     * // data is inside `rows` array:
     * result.rows[0].one;
     * ```
     *
     * If the query is executing inside a transaction, it will use the transaction connection automatically.
     *
     * ```ts
     * await db.transaction(async () => {
     *   // both queries will execute in the same transaction
     *   await db.$query`SELECT 1`;
     *   await db.$query`SELECT 2`;
     * });
     * ```
     *
     * Alternatively, support a simple SQL string, with optional `values`:
     *
     * Note that the values is a simple array, and the SQL is referring to the values with `$1`, `$2` and so on.
     *
     * ```ts
     * const value = 1;
     *
     * // it is NOT safe to interpolate inside a simple string, use `values` to pass the values.
     * const result = await db.$query<{ one: number }>({
     *   raw: 'SELECT $1 AS one',
     *   values: [value],
     * });
     * // data is inside `rows` array:
     * result.rows[0].one;
     * ```
     *
     * @param args - SQL template literal, or an object { raw: string, values?: unknown[] }
     */
    $query: Db['query'];
    /**
     * The same as the {@link $query}, but returns an array of arrays instead of objects:
     *
     * ```ts
     * const value = 1;
     *
     * // it is safe to interpolate inside the backticks (``):
     * const result = await db.$queryArrays<[number]>`SELECT ${value} AS one`;
     * // `rows` is an array of arrays:
     * const row = result.rows[0];
     * row[0]; // our value
     * ```
     *
     * @param args - SQL template literal, or an object { raw: string, values?: unknown[] }
     */
    $queryArrays: Db['queryArrays'];
    /**
     * See {@link FromMethods.from}
     */
    $from<Arg extends MaybeArray<FromArg<Query>>>(arg: Arg): FromResult<Query, Arg>;
    $close(): Promise<void>;
};
type OrchidOrmArg = true | null extends true ? 'Set strict: true to tsconfig' : ({
    db: Query;
} | {
    adapter: Adapter;
} | Omit<AdapterOptions, 'log'>) & DbSharedOptions;
declare const orchidORM: <T extends TableClasses>({ log, logger, autoPreparedStatements, noPrimaryKey, ...options }: OrchidOrmArg, tables: T) => OrchidORM<T>;

interface HasAndBelongsToMany extends RelationThunkBase {
    type: 'hasAndBelongsToMany';
    options: HasAndBelongsToManyOptions;
}
type HasAndBelongsToManyOptions<Columns extends ColumnsShapeBase = ColumnsShapeBase, Related extends TableClass = TableClass, Scope extends Query = Query> = RelationCommonOptions<Related, Scope> & ({
    columns: (keyof Columns)[];
    references: string[];
    through: {
        table: string;
        columns: string[];
        references: (keyof InstanceType<Related>['columns']['shape'])[];
    };
} | {
    primaryKey: keyof Columns;
    foreignKey: string;
    joinTable: string;
    associationPrimaryKey: string;
    associationForeignKey: keyof InstanceType<Related>['columns']['shape'];
});
type HasAndBelongsToManyParams<T extends RelationConfigSelf, Relation extends RelationThunkBase> = Relation['options'] extends {
    columns: string[];
} ? {
    [Name in Relation['options']['columns'][number]]: T['columns']['shape'][Name]['type'];
} : Relation['options'] extends {
    primaryKey: string;
} ? Record<Relation['options']['primaryKey'], T['columns']['shape'][Relation['options']['primaryKey']]['type']> : never;
interface HasAndBelongsToManyInfo<T extends RelationConfigSelf, Name extends keyof T['relations'] & string, TableQuery extends Query, Q extends Query = {
    [K in keyof TableQuery]: K extends 'meta' ? Omit<TableQuery['meta'], 'selectable'> & {
        as: Name;
        hasWhere: true;
        selectable: SelectableFromShape<TableQuery['shape'], Name>;
    } : K extends 'join' ? RelJoin : TableQuery[K];
}> {
    query: Q;
    methodQuery: Q;
    joinQuery: RelationJoinQuery;
    one: false;
    omitForeignKeyInCreate: never;
    optionalDataForCreate: {
        [P in Name]?: RelationToManyDataForCreate<{
            nestedCreateQuery: Q;
            table: Q;
        }>;
    };
    dataForUpdate: {
        disconnect?: MaybeArray<WhereArg<Q>>;
        set?: MaybeArray<WhereArg<Q>>;
        delete?: MaybeArray<WhereArg<Q>>;
        update?: {
            where: MaybeArray<WhereArg<Q>>;
            data: UpdateData<Q>;
        };
        create?: CreateData<Q>[];
    };
    dataForUpdateOne: EmptyObject;
    params: HasAndBelongsToManyParams<T, T['relations'][Name]>;
}

type RelationToOneDataForCreate<Rel extends {
    nestedCreateQuery: Query;
    table: Query;
}> = {
    create: CreateData<Rel['nestedCreateQuery']>;
    connect?: never;
    connectOrCreate?: never;
} | {
    create?: never;
    connect: WhereArg<Rel['table']>;
    connectOrCreate?: never;
} | {
    create?: never;
    connect?: never;
    connectOrCreate: {
        where: WhereArg<Rel['table']>;
        create: CreateData<Rel['nestedCreateQuery']>;
    };
};
type RelationToOneDataForCreateSameQuery<Q extends Query> = {
    create: CreateData<Q>;
    connect?: never;
    connectOrCreate?: never;
} | {
    create?: never;
    connect: WhereArg<Q>;
    connectOrCreate?: never;
} | {
    create?: never;
    connect?: never;
    connectOrCreate: {
        where: WhereArg<Q>;
        create: CreateData<Q>;
    };
};
type RelationToManyDataForCreate<Rel extends {
    nestedCreateQuery: Query;
    table: Query;
}> = {
    create?: CreateData<Rel['nestedCreateQuery']>[];
    connect?: WhereArg<Rel['table']>[];
    connectOrCreate?: {
        where: WhereArg<Rel['table']>;
        create: CreateData<Rel['nestedCreateQuery']>;
    }[];
};
interface RelationThunkBase {
    type: string;
    fn(): TableClass;
    options: RelationCommonOptions;
}
type RelationThunk = BelongsTo | HasOne | HasMany | HasAndBelongsToMany;
interface RelationThunks {
    [K: string]: RelationThunk;
}
type RelationScopeOrTable<Relation extends RelationThunkBase> = Relation['options']['scope'] extends (q: Query) => Query ? ReturnType<Relation['options']['scope']> : DbTable<InstanceType<ReturnType<Relation['fn']>>>;
interface RelationConfigSelf {
    columns: {
        shape: ColumnsShapeBase;
    };
    relations: RelationThunks;
}
type RelationConfigParams<T extends RelationConfigSelf, Relation extends RelationThunk> = Relation extends BelongsTo ? BelongsToParams<T, Relation> : Relation extends HasOne ? HasOneParams<T, Relation> : Relation extends HasMany ? HasManyParams<T, Relation> : Relation extends HasAndBelongsToMany ? HasAndBelongsToManyParams<T, Relation> : never;
type MapRelation<T extends RelationConfigSelf, K extends keyof T['relations'] & string> = RelationQuery<T['relations'][K] extends BelongsTo ? BelongsToInfo<T, K, RelationScopeOrTable<T['relations'][K]>, BelongsToFKey<T['relations'][K]>, T['relations'][K]['options']['required']> : T['relations'][K] extends HasOne ? HasOneInfo<T, K, RelationScopeOrTable<T['relations'][K]>> : T['relations'][K] extends HasMany ? HasManyInfo<T, K, RelationScopeOrTable<T['relations'][K]>> : T['relations'][K] extends HasAndBelongsToMany ? HasAndBelongsToManyInfo<T, K, RelationScopeOrTable<T['relations'][K]>> : never>;
type MapRelations<T> = T extends RelationConfigSelf ? {
    [K in keyof T['relations'] & string]: MapRelation<T, K>;
} : EmptyObject;

interface TableClass<T extends Table = Table> {
    new (): T;
    instance(): T;
}
interface TableClasses {
    [K: string]: TableClass;
}
interface RelationQueriesBase {
    [K: string]: RelationQueryBase;
}
interface TableToDb<T extends Table, RelationQueries extends RelationQueriesBase> extends Db<T['table'], T['columns']['shape'], keyof ShapeColumnPrimaryKeys<T['columns']['shape']> extends never ? never : ShapeColumnPrimaryKeys<T['columns']['shape']>, ShapeUniqueColumns<T['columns']['shape']> | TableDataItemsUniqueColumns<T['columns']['shape'], T['columns']['data']>, TableDataItemsUniqueColumnTuples<T['columns']['shape'], T['columns']['data']>, UniqueConstraints<T['columns']['shape']> | TableDataItemsUniqueConstraints<T['columns']['data']>, RelationQueries, T['types'], T['computed'] extends RecordUnknown ? T['columns']['shape'] & {
    [K in keyof T['computed']]: ReturnType<T['computed'][K]>['result']['value'];
} : T['columns']['shape'], MapTableScopesOption<T['scopes'], T['softDelete']>> {
    definedAs: string;
    db: OrchidORM;
    getFilePath(): string;
    name: string;
}
type DbTable<T extends Table> = TableToDb<T, MapRelations<T>> & MapRelations<T>;
type ScopeFn<Related extends TableClass, Scope extends Query> = (q: DbTable<InstanceType<Related>>) => Scope;
interface Table {
    table: string;
    columns: {
        shape: ColumnsShapeBase;
        data: MaybeArray<TableDataItem>;
    };
    schema?: string;
    types: unknown;
    noPrimaryKey?: boolean;
    filePath: string;
    language?: string;
    /**
     * collect computed columns returned by {@link BaseTable.setColumns}
     */
    computed?: ComputedColumnsBase<never>;
    scopes?: CoreQueryScopes;
    readonly softDelete?: true | string;
    comment?: string;
}
type Queryable<T extends Table> = {
    [K in keyof T['columns']['shape']]?: T['columns']['shape'][K]['queryType'];
};
type Selectable<T extends Table> = ColumnShapeOutput<T['columns']['shape']>;
type Insertable<T extends Table> = ColumnShapeInput<T['columns']['shape']>;
type Updatable<T extends Table> = ColumnShapeInputPartial<T['columns']['shape']>;
type BeforeHookMethod = <T extends Table>(cb: QueryBeforeHook) => T;
type AfterHookMethod = <T extends Table>(cb: QueryAfterHook) => T;
type AfterSelectableHookMethod = <T extends Table, S extends (keyof T['columns']['shape'])[]>(this: T, select: S, cb: AfterHook<S, T['columns']['shape']>) => T;
interface SetColumnsResult<Shape extends ColumnsShapeBase, Data extends MaybeArray<MaybeArray<TableDataItem>>> {
    shape: Shape;
    data: Data extends unknown[] ? Data : [Data];
}
interface BaseTableInstance<ColumnTypes> {
    table: string;
    columns: {
        shape: ColumnsShapeBase;
        data: MaybeArray<TableDataItem>;
    };
    schema?: string;
    noPrimaryKey?: boolean;
    snakeCase?: boolean;
    types: ColumnTypes;
    q: QueryData;
    language?: string;
    filePath: string;
    result: ColumnsShapeBase;
    clone<T extends QueryBase>(this: T): T;
    getFilePath(): string;
    setColumns<Shape extends ColumnsShapeBase, Data extends MaybeArray<TableDataItem>>(fn: (t: ColumnTypes) => Shape, tableData?: TableDataFn<Shape, Data>): SetColumnsResult<Shape, Data>;
    /**
     * You can add a generated column in the migration (see [generated](/guide/migration-column-methods.html#generated-column)),
     * such column will persist in the database, it can be indexed.
     *
     * Or you can add a computed column on the ORM level, without adding it to the database, in such a way:
     *
     * ```ts
     * import { BaseTable } from './baseTable';
     *
     * export class UserTable extends BaseTable {
     *   readonly table = 'user';
     *   columns = this.setColumns((t) => ({
     *     id: t.identity().primaryKey(),
     *     firstName: t.string(),
     *     lastName: t.string(),
     *   }));
     *
     *   computed = this.setComputed({
     *     fullName: (q) =>
     *       q.sql`${q.column('firstName')} || ' ' || ${q.column('lastName')}`.type(
     *         (t) => t.string(),
     *       ),
     *   });
     * }
     * ```
     *
     * `setComputed` takes an object where keys are computed column names, and values are functions returning raw SQL.
     *
     * Use `q.column` as shown above to reference a table column, it will be prefixed with a correct table name even if the table is joined under a different name.
     *
     * Computed columns are not selected by default, only on demand:
     *
     * ```ts
     * const a = await db.user.take();
     * a.fullName; // not selected
     *
     * const b = await db.user.select('*', 'fullName');
     * b.fullName; // selected
     *
     * // Table post belongs to user as an author.
     * // it's possible to select joined computed column:
     * const posts = await db.post
     *   .join('author')
     *   .select('post.title', 'author.fullName');
     * ```
     *
     * SQL query can be generated dynamically based on the current request context.
     *
     * Imagine we are using [AsyncLocalStorage](https://nodejs.org/api/async_context.html#asynchronous-context-tracking)
     * to keep track of current user's language.
     *
     * And we have articles translated to different languages, each article has `title_en`, `title_uk`, `title_be` and so on.
     *
     * We can define a computed `title` by passing a function into `sql` method:
     *
     * ```ts
     * type Locale = 'en' | 'uk' | 'be';
     * const asyncLanguageStorage = new AsyncLocalStorage<Locale>();
     * const defaultLocale: Locale = 'en';
     *
     * export class ArticleTable extends BaseTable {
     *   readonly table = 'article';
     *   columns = this.setColumns((t) => ({
     *     id: t.identity().primaryKey(),
     *     title_en: t.text(),
     *     title_uk: t.text().nullable(),
     *     title_be: t.text().nullable(),
     *   }));
     *
     *   computed = this.setComputed({
     *     title: (q) =>
     *       q
     *         // .sql can take a function that accepts `sql` argument and must return SQL
     *         .sql((sql) => {
     *           // get locale dynamically based on current storage value
     *           const locale = asyncLanguageStorage.getStore() || defaultLocale;
     *
     *           // use COALESCE in case when localized title is NULL, use title_en
     *           return sql`COALESCE(
     *             ${q.column(`title_${locale}`)},
     *             ${q.column(`title_${defaultLocale}`)}
     *           )`;
     *         })
     *         .type((t) => t.text()),
     *   });
     * }
     * ```
     *
     * @param computed - object where keys are column names and values are functions returning raw SQL
     */
    setComputed<Table extends string, Shape extends ColumnsShapeBase, Computed extends ComputedColumnsBase<Db<Table, Shape, never, never, never, never, EmptyObject, ColumnTypes>>>(computed: Computed): Computed;
    /**
     * See {@link ScopeMethods}
     */
    setScopes<Table extends string, Columns extends ColumnsShapeBase, Keys extends string>(this: {
        table: Table;
        columns: {
            shape: Columns;
        };
    }, scopes: DbTableOptionScopes<Table, Columns, Keys>): CoreQueryScopes<Keys>;
    belongsTo<Columns extends ColumnsShapeBase, Related extends TableClass, Scope extends Query, Options extends BelongsToOptions<Columns, Related, Scope>>(this: {
        columns: {
            shape: Columns;
        };
    }, fn: () => Related, options: Options): {
        type: 'belongsTo';
        fn: () => Related;
        options: Options;
    };
    hasOne<Columns extends ColumnsShapeBase, Related extends TableClass, Scope extends Query, Through extends string, Source extends string, Options extends HasOneOptions<Columns, Related, Scope, Through, Source>>(this: {
        columns: {
            shape: Columns;
        };
    }, fn: () => Related, options: Options): {
        type: 'hasOne';
        fn: () => Related;
        options: Options;
    };
    hasMany<Columns extends ColumnsShapeBase, Related extends TableClass, Scope extends Query, Through extends string, Source extends string, Options extends HasManyOptions<Columns, Related, Scope, Through, Source>>(this: {
        columns: {
            shape: Columns;
        };
    }, fn: () => Related, options: Options): {
        type: 'hasMany';
        fn: () => Related;
        options: Options;
    };
    hasAndBelongsToMany<Columns extends ColumnsShapeBase, Related extends TableClass, Scope extends Query, Options extends HasAndBelongsToManyOptions<Columns, Related, Scope>>(this: {
        columns: {
            shape: Columns;
        };
    }, fn: () => Related, options: Options): {
        type: 'hasAndBelongsToMany';
        fn: () => Related;
        options: Options;
    };
    beforeQuery: BeforeHookMethod;
    afterQuery: AfterHookMethod;
    beforeCreate: BeforeHookMethod;
    afterCreate: AfterSelectableHookMethod;
    afterCreateCommit: AfterSelectableHookMethod;
    beforeUpdate: BeforeHookMethod;
    afterUpdate: AfterSelectableHookMethod;
    afterUpdateCommit: AfterSelectableHookMethod;
    beforeSave: BeforeHookMethod;
    afterSave: AfterSelectableHookMethod;
    afterSaveCommit: AfterSelectableHookMethod;
    beforeDelete: BeforeHookMethod;
    afterDelete: AfterSelectableHookMethod;
    afterDeleteCommit: AfterSelectableHookMethod;
}
interface BaseTableClass<SchemaConfig extends ColumnSchemaConfig, ColumnTypes> extends SqlMethod<ColumnTypes> {
    nowSQL: string | undefined;
    exportAs: string;
    columnTypes: ColumnTypes;
    getFilePath(): string;
    new (): BaseTableInstance<ColumnTypes>;
    instance(): BaseTableInstance<ColumnTypes>;
    inputSchema: SchemaConfig['inputSchema'];
    outputSchema: SchemaConfig['outputSchema'];
    querySchema: SchemaConfig['querySchema'];
}
declare function createBaseTable<SchemaConfig extends ColumnSchemaConfig = DefaultSchemaConfig, ColumnTypes = DefaultColumnTypes<SchemaConfig>>({ schemaConfig, columnTypes: columnTypesArg, snakeCase, filePath: filePathArg, nowSQL, exportAs, language, }?: {
    schemaConfig?: SchemaConfig;
    columnTypes?: ColumnTypes | ((t: DefaultColumnTypes<SchemaConfig>) => ColumnTypes);
    snakeCase?: boolean;
    filePath?: string;
    nowSQL?: string;
    exportAs?: string;
    language?: string;
}): BaseTableClass<SchemaConfig, ColumnTypes>;

type QueryMethods<T extends Query> = Record<string, (q: T, ...args: any[]) => any>;
type QueryOne<T extends Query> = {
    [K in keyof T]: K extends 'returnType' ? Exclude<QueryReturnType, 'all'> : T[K];
};
interface MethodsBase<T extends Query> {
    queryMethods?: QueryMethods<T>;
    queryOneMethods?: QueryMethods<QueryOne<T>>;
    queryWithWhereMethods?: QueryMethods<WhereResult<T>>;
    queryOneWithWhereMethods?: QueryMethods<QueryOne<WhereResult<T>>>;
    methods?: RecordUnknown;
}
type MapQueryMethods<BaseQuery extends Query, Method> = Method extends (q: any, ...args: infer Args) => infer Result ? <T extends BaseQuery>(this: T, ...args: Args) => Result extends Query ? MergeQuery<T, Result> : Result : never;
type MapMethods<T extends Query, Methods extends MethodsBase<T>> = {
    [K in keyof Methods['queryMethods'] | keyof Methods['queryOneMethods'] | keyof Methods['queryWithWhereMethods'] | keyof Methods['queryOneWithWhereMethods'] | keyof Methods['methods']]: K extends keyof Methods['methods'] ? Methods['methods'][K] : K extends keyof Methods['queryOneWithWhereMethods'] ? MapQueryMethods<QueryOne<WhereResult<Query>>, Methods['queryOneWithWhereMethods'][K]> : K extends keyof Methods['queryWithWhereMethods'] ? MapQueryMethods<WhereResult<Query>, Methods['queryWithWhereMethods'][K]> : K extends keyof Methods['queryOneMethods'] ? MapQueryMethods<QueryOne<Query>, Methods['queryOneMethods'][K]> : K extends keyof Methods['queryMethods'] ? MapQueryMethods<Query, Methods['queryMethods'][K]> : never;
};
type Repo<T extends Query, Methods extends MethodsBase<T>> = (<Q extends {
    table: T['table'];
    shape: T['shape'];
}>(q: Q) => Query & Q & MapMethods<T, Methods>) & T & MapMethods<T, Methods>;
declare const createRepo: <T extends Query, Methods extends MethodsBase<T>>(table: T, methods: Methods) => Repo<T, Methods>;

export { BaseTableClass, BaseTableInstance, DbTable, Insertable, MapMethods, MapQueryMethods, MethodsBase, OrchidORM, Queryable, Repo, ScopeFn, Selectable, SetColumnsResult, Table, TableClass, TableClasses, TableToDb, Updatable, createBaseTable, createRepo, orchidORM };
